/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: LibvirtServerPool.java 191 2011-07-25 13:43:26Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.libvirt;

import java.util.Map;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractStoragePool;
import org.ow2.sirocco.vmm.agent.domain.ServerPool;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;

public class LibvirtServerPool extends ServerPool {
    static Logger logger = Logger.getLogger(LibvirtServerPool.class);

    static long DEFAULT_HOST_SYNC_PERIOD_MILLIS = 10000;

    private RemoteExec.SshAuthInfo defaultAuthInfo;

    private String hypervisor;

    private long syncPeriodMillis = LibvirtServerPool.DEFAULT_HOST_SYNC_PERIOD_MILLIS;

    public LibvirtServerPool(final String name, final ObjectName objectName, final Map<String, String> attributes)
        throws IllegalArgumentException {
        super(name, objectName, attributes);
        String sshPassword = attributes.get("sshRootPassword");
        String sshPrivateKeyFile = attributes.get("sshPrivateKeyFile");
        if (sshPassword != null || sshPrivateKeyFile != null) {
            this.defaultAuthInfo = new RemoteExec.SshAuthInfo(sshPassword, sshPrivateKeyFile);
        }
        this.hypervisor = attributes.get("hypervisor");
        if (this.hypervisor == null) {
            throw new IllegalArgumentException("Missing parameter: hypervisor");
        }
        String period = attributes.get("syncPeriodMillis");
        if (period != null) {
            try {
                this.syncPeriodMillis = Integer.parseInt(period);
            } catch (NumberFormatException ex) {
                throw new IllegalArgumentException("Wrong value for parameter syncPeriodMillis");
            }
        }
    }

    @Override
    public String getHypervisor() {
        return this.hypervisor;
    }

    @Override
    public HostMXBean newHost(final String hostName, final Map<String, String> props) throws VMMException {
        try {
            for (HostMXBean host : this.getManagedHosts()) {
                if (host.getHostName().equals(hostName)) {
                    return host;
                }
            }
            ObjectName mbeanObjectName = MBeanObjectNamer.makeHostName(this.getPath() + "/" + hostName, hostName);
            RemoteExec.SshAuthInfo hostAuthInfo;
            if (props.get("sshRootPassword") != null) {
                hostAuthInfo = new RemoteExec.SshAuthInfo(props.get("sshRootPassword"), null);
            } else if (props.get("sshPrivateKeyFile") != null) {
                hostAuthInfo = new RemoteExec.SshAuthInfo(null, props.get("sshPrivateKeyFile"));
            } else {
                hostAuthInfo = this.defaultAuthInfo;
            }
            if (hostAuthInfo == null) {
                throw new IllegalArgumentException("Missing SSH authentification info for host " + hostName);
            }

            String url = props.get("url");
            if (url == null) {
                throw new IllegalArgumentException("Missing parameter: url");
            }

            LibvirtHost host = LibvirtHost.newHost(this, mbeanObjectName, hostName, url, hostAuthInfo, props, true);
            if (host != null) {
                AgentCommon.getMBeanServer().registerMBean(host, mbeanObjectName);
                this.addHost(host);
                LibvirtServerPool.logger
                    .info("In server pool " + this.serverPoolName + ": added Host MBean " + mbeanObjectName);

            } else {
                host = LibvirtHost.newHost(this, mbeanObjectName, hostName, url, hostAuthInfo, props, false);
                if (host != null) {
                    AgentCommon.getMBeanServer().registerMBean(host, mbeanObjectName);
                    this.addHost(host);
                    LibvirtServerPool.logger.info("In server pool " + this.serverPoolName + ": added sleeping Host MBean "
                        + mbeanObjectName);
                }
            }

            return host;
        } catch (Exception ex) {
            LibvirtServerPool.logger.error("Failed to add LibvirtHost", ex);
            throw new VMMException(ex);
        }
    }

    @Override
    protected AbstractStoragePool newStoragePool(final String name, final ObjectName objectName,
        final Map<String, String> attributes) throws Exception {
        if (this.getManagedHosts().size() == 0) {
            throw new VMMException("Cannot add libvirt storage pool without a host");
        }
        // XXX all storage pools are attatched to the first host of the server
        // pool
        LibvirtHost host = (LibvirtHost) this.getManagedHosts().get(0);
        return new LibvirtStoragePool(objectName, name, host, attributes);
    }

    @Override
    public VirtualMachineImageStoreMXBean getVMImageStore() {
        return null;
    }

    public synchronized long getSyncPeriodMillis() {
        return this.syncPeriodMillis;
    }

    @Override
    public void deleteImageStore() {
    }

}
