/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: LibvirtVirtualMachine.java 710 2011-12-08 17:38:25Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.libvirt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.ObjectName;
import javax.xml.xpath.XPathConstants;

import org.apache.log4j.Logger;
import org.libvirt.Domain;
import org.libvirt.DomainInfo;
import org.libvirt.DomainInfo.DomainState;
import org.libvirt.LibvirtException;
import org.libvirt.SchedParameter;
import org.libvirt.SchedUintParameter;
import org.libvirt.VcpuInfo;
import org.ow2.sirocco.vmm.agent.domain.AbstractVirtualMachine;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.BadVMPowerStateException;
import org.ow2.sirocco.vmm.api.DiskOperation;
import org.ow2.sirocco.vmm.api.GuestInfo;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.IllegalOperationException;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourceUsage;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMCustomizationSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VNIC;
import org.ow2.sirocco.vmm.api.VirtualCdrom;
import org.ow2.sirocco.vmm.api.VirtualDisk;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.Volume;
import org.w3c.dom.Node;

/**
 * Libvirt driver Virtual Machine MXBean implementation
 */
public class LibvirtVirtualMachine extends AbstractVirtualMachine implements VirtualMachineMXBean {
    static Logger logger = Logger.getLogger(LibvirtVirtualMachine.class);

    // separator used between tags in libvirt XML description element
    static String TAG_SEPARATOR = ";";

    private LibvirtHost host;

    private Domain domain;

    private String uuid;

    private String name;

    private String ipAddress;

    private String macAddress;

    private PowerState cachedPowerState = PowerState.UNKNOWN;

    boolean isSynchronized;

    private boolean isMigrating;

    private List<VirtualDisk> disks;

    private List<VirtualCdrom> cdroms;

    private Map<String, String> userMetadata = new HashMap<String, String>();

    public LibvirtVirtualMachine(final ObjectName objectName, final LibvirtHost host, final Domain domain,
        final Map<String, String> userData) throws VMMException {
        super(objectName);
        this.isSynchronized = true;
        try {
            this.host = host;
            this.domain = domain;
            this.uuid = domain.getUUIDString();
            this.name = domain.getName();
            this.initUserMetadata();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    private void initUserMetadata() throws LibvirtException {
        String description = LibvirtConfigFileUtils.getDescription(this.domain.getXMLDesc(0));
        String tags[] = description.split(LibvirtVirtualMachine.TAG_SEPARATOR);
        for (String tag : tags) {
            String[] keyValue = tag.split("=", 2);
            if (keyValue.length >= 2) {
                this.userMetadata.put(keyValue[0], keyValue[1]);
            }
        }
    }

    void release() {
        try {
            if (this.domain != null) {
                this.domain.free();
            }
        } catch (LibvirtException e) {
        }
        this.domain = null;
    }

    public boolean canLiveMigrateToHost(final HostMXBean targetHost) throws VMMException {
        LibvirtHost host = null;
        for (HostMXBean h : this.host.getServerPool().getManagedHosts()) {
            if (h.getObjectName().equals(targetHost.getObjectName())) {
                host = (LibvirtHost) h;
                break;
            }
        }
        return host != null && host != this.host && this.getMemorySizeMB() < host.getFreeMemoryMB();
    }

    public synchronized boolean isSynchronized() {
        return this.isSynchronized;
    }

    public synchronized void setSynchronized(final boolean isSynchronized) {
        this.isSynchronized = isSynchronized;
    }

    public synchronized boolean isMigrating() {
        return this.isMigrating;
    }

    public synchronized void setMigrating(final boolean isMigrating) {
        this.isMigrating = isMigrating;
    }

    public VirtualMachineMXBean cloneVM(final String clonedVmName, final VMCustomizationSpec custSpec, final boolean sync)
        throws InsufficientResourcesException, VMMException {
        return this.host.cloneVM(this.getNameLabel(), clonedVmName, custSpec, sync);
    }

    public String getUuid() {
        return this.uuid;
    }

    public HostMXBean getHostMBean() {
        return this.host;
    }

    public long getDomID() {
        try {
            return this.domain.getID();
        } catch (LibvirtException ex) {
            LibvirtVirtualMachine.logger.error(ex);
            return -1;
        }
    }

    public List<String> getDiskFiles() throws VMMException {
        try {
            String xmlDesc = this.domain.getXMLDesc(0);
            return LibvirtConfigFileUtils.getDiskFiles(xmlDesc);
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public Date getStartTime() {
        return null;
    }

    public long getUpTimeSeconds() {
        return -1;
    }

    private ResourceUsage currentResourceUsage = new ResourceUsage();

    void updateCurrentResourceUsage(final ResourceUsage usage) {
        this.currentResourceUsage = usage;
    }

    public ResourceUsage getResourceUsage() {
        return this.currentResourceUsage;
    }

    public float getCPULoad() {
        return this.currentResourceUsage.getCpuLoad();
    }

    public float[] getLoadPerVCPU() throws VMMException {
        try {
            VcpuInfo vcpuInfos1[] = this.domain.getVcpusInfo();
            try {
                Thread.sleep(500);
            } catch (Exception ex) {
            }
            VcpuInfo vcpuInfos2[] = this.domain.getVcpusInfo();
            float vcpuLoad[] = new float[vcpuInfos1.length];
            for (int i = 0; i < vcpuInfos1.length; i++) {
                vcpuLoad[i] = (float) ((vcpuInfos2[i].cpuTime - vcpuInfos1[i].cpuTime) / Math.pow(10, 9));
            }
            return vcpuLoad;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public long getMemorySizeMB() throws VMMException {
        try {
            return this.domain.getInfo().maxMem / 1024;
        } catch (LibvirtException ex) {
            throw new VMMException(ex);
        }
    }

    public void setMemorySizeMB(final long size) throws VMMException {
        try {
            this.domain.setMaxMemory(size * 1024);
            this.domain.setMemory(size * 1024);
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public long getMemoryUsedMB() throws VMMException {
        try {
            return this.domain.getInfo().memory / 1024;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public String getNameLabel() {
        return this.name;
    }

    @Override
    public Map<String, String> getUserMetadata() {
        return this.userMetadata;
    }

    @Override
    public void createUserMetadata(final Map<String, String> metadata) throws VMMException {
        this.userMetadata.putAll(metadata);
        try {
            LibvirtDomainConfig domainConfig = LibvirtDomainConfig.loadDomainConfig(this.domain.getXMLDesc(0));
            StringBuffer description = new StringBuffer();
            for (Map.Entry<String, String> tag : this.userMetadata.entrySet()) {
                description.append(tag.getKey() + "=" + tag.getValue() + LibvirtVirtualMachine.TAG_SEPARATOR);
            }
            domainConfig.setDescription(description.toString());
            this.domain.getConnect().domainDefineXML(domainConfig.toString());
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }

    }

    public String getConsole() throws VMMException {
        try {
            String xmlDesc = this.domain.getXMLDesc(0);
            String vncPortString = (String) LibvirtConfigFileUtils.getConfigItem(xmlDesc,
                "/domain/devices/graphics[@type='vnc']/@port", XPathConstants.STRING);
            if (vncPortString == null || vncPortString.equals("")) {
                return null;
            }
            String console = "vnc://" + this.host.getHostName() + ":" + Integer.parseInt(vncPortString);
            LibvirtVirtualMachine.logger.debug("VM " + this.name + " console=" + console);
            return console;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getNumVCPUs()
     */
    public int getNumVCPUs() throws VMMException {
        DomainInfo info;
        try {
            info = this.domain.getInfo();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
        return info.nrVirtCpu;
    }

    @Override
    public boolean[][] getCPUAffinity() throws VMMException {
        try {
            int numVCPUs = this.domain.getInfo().nrVirtCpu;
            int numPCPUs = this.domain.getConnect().nodeInfo().cpus;
            int numMaxPCPUs = this.domain.getConnect().nodeInfo().maxCpus();
            int[] cpuMap = this.domain.getVcpusCpuMaps();
            boolean[][] result = new boolean[numVCPUs][];
            for (int i = 0; i < numVCPUs; i++) {
                result[i] = new boolean[numPCPUs];
            }
            int maxVCPUS = this.domain.getMaxVcpus();
            int cpuMapLength = this.domain.cpuMapLength(numMaxPCPUs);
            for (int i = 0; i < numVCPUs; i++) {
                for (int j = 0; j < numPCPUs; j++) {
                    result[i][j] = (cpuMap[i * cpuMapLength + (j / 8)] & (1 << (j % 8))) != 0;
                }
            }

            return result;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    @Override
    public void setCPUAffinity(final boolean[][] affinity) throws VMMException {
        try {
            if (affinity == null) {
                throw new IllegalArgumentException();
            }
            int mapLength = (this.domain.getConnect().nodeInfo().cpus + 7) / 8;
            for (int vcpuIndex = 0; vcpuIndex < affinity.length; vcpuIndex++) {
                int[] cpuMap = new int[mapLength];
                if (affinity[vcpuIndex] == null) {
                    throw new IllegalArgumentException();
                }
                for (int pcpuIndex = 0; pcpuIndex < affinity[vcpuIndex].length; pcpuIndex++) {
                    if (affinity[vcpuIndex][pcpuIndex]) {
                        cpuMap[pcpuIndex / 8] |= (1 << (pcpuIndex % 8));
                    }
                }
                this.domain.pinVcpu(vcpuIndex, cpuMap);
            }
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getSchedulingCap()
     */
    public int getSchedulingCap() throws VMMException {
        try {
            SchedParameter[] schedParams = this.domain.getSchedulerParameters();
            for (SchedParameter param : schedParams) {
                if (param.field.equals("cap")) {
                    return ((SchedUintParameter) param).value;
                }
            }
            return -1;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getSchedulingWeight()
     */
    public int getSchedulingWeight() throws VMMException {
        try {
            SchedParameter[] schedParams = this.domain.getSchedulerParameters();
            for (SchedParameter param : schedParams) {
                if (param.field.equals("weight")) {
                    return ((SchedUintParameter) param).value;
                }
            }
            return -1;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getState()
     */
    public PowerState getState() throws VMMException {
        DomainState state;
        try {
            state = this.domain.getInfo().state;
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }

        PowerState newState = PowerState.UNKNOWN;
        switch (state) {
        case VIR_DOMAIN_BLOCKED:
        case VIR_DOMAIN_RUNNING:
        case VIR_DOMAIN_SHUTDOWN:
            newState = PowerState.RUNNING;
            break;
        case VIR_DOMAIN_SHUTOFF:
            newState = PowerState.HALTED;
            break;
        case VIR_DOMAIN_PAUSED:
            newState = PowerState.PAUSED;
            break;
        case VIR_DOMAIN_CRASHED:
            newState = PowerState.UNKNOWN;
            break;
        case VIR_DOMAIN_NOSTATE:
            // XXX virDomainGetInfo sometimes returns dubious NOSTATE status for
            // running Xen guests
            newState = this.cachedPowerState;
            break;
        }
        if (newState == PowerState.UNKNOWN) {
            LibvirtVirtualMachine.logger.debug("VM " + this.getNameLabel() + " state=" + state);
        }
        if (newState != this.cachedPowerState) {
            this.emitNotification(NotificationTypes.VM_STATE_CHANGE, newState.toString(), null);
            LibvirtVirtualMachine.logger.debug("VM " + this.getNameLabel() + " new power state: " + newState.toString());
            this.cachedPowerState = newState;
        }
        return this.cachedPowerState;
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getGuestIpAddress()
     */
    public String getGuestIpAddress() throws VMMException {
        try {
            if (this.getState() != PowerState.RUNNING) {
                return "unknown";
            }
        } catch (Exception ex) {
        }
        if (this.ipAddress == null) {
            // wait until the guest OS has booted to acquire its IP address
            int nbAttempts = 10;
            while (nbAttempts-- > 0) {
                try {
                    this.ipAddress = this.host.getIPAddress(this.getMacAddress());
                } catch (Exception ex) {
                    throw new VMMException(ex.getMessage());
                }
                if (this.ipAddress != null && !this.ipAddress.equals("")) {
                    break;
                }
                try {
                    TimeUnit.SECONDS.sleep(10);
                } catch (InterruptedException ex) {
                }
            }
        }
        return this.ipAddress == null ? "unknown" : this.ipAddress;
    }

    @Override
    public GuestInfo getGuestInfo() throws VMMException {
        String script = AgentCommon.getScript(LibvirtDriver.NAME, LibvirtDriver.GET_IP_FROM_MAC_SCRIPT);
        StringBuilder command = new StringBuilder("bash -s ");
        command.append(this.name + " " + this.getMacAddress());
        command.append(" << 'END'\n");
        command.append(script);
        command.append("\n");
        command.append("END");

        // LibvirtVirtualMachine.logger.debug("Launching command " +
        // command.toString());

        RemoteExec.Result result;
        try {
            result = RemoteExec.commandAsRoot(this.host.getHostName(), this.host.getSshAuthInfo(), command.toString());
        } catch (RemoteExec.SshException ex) {
            LibvirtHost.logger.debug("getGuestInfo failed ", ex);
            throw new VMMException("Internal error: " + ex.getMessage());
        }
        if (result.exitCode != 0) {
            LibvirtHost.logger.error("getIPfromMac script failed: " + result.output);
            throw new VMMException("Internal error: " + result.error);
        }
        if (result.output.indexOf('.') != -1) {
            GuestInfo guestInfo = new GuestInfo();
            String ip = result.output;
            if (ip.endsWith("\n")) {
                ip = ip.substring(0, ip.length() - 1);
            }
            guestInfo.setIpAddresses(Collections.singletonList(ip));
            return guestInfo;
        }
        /*
         * int i = result.output.indexOf(' '); if (i != -1) { int j =
         * result.output.indexOf(' ', i + 1); if (j != -1) { String ip =
         * result.output.substring(0, i); String hostName =
         * result.output.substring(i + 1, j); String osType =
         * result.output.substring(j + 1).trim(); GuestInfo guestInfo = new
         * GuestInfo(); guestInfo.setHostName(hostName);
         * guestInfo.setOsType(osType);
         * guestInfo.setIpAddresses(Collections.singletonList(ip)); return
         * guestInfo; } }
         */
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#getMacAddress()
     */
    public String getMacAddress() throws VMMException {
        if (this.macAddress == null) {
            String xmlDesc;
            try {
                xmlDesc = this.domain.getXMLDesc(0);
            } catch (LibvirtException ex) {
                throw LibvirtVirtualMachine.translateLibvirtException(ex);
            }
            int i = xmlDesc.indexOf("mac address");
            if (i == -1) {
                return null;
            }
            int startIndex = i + "mac address".length() + 2;
            this.macAddress = xmlDesc.substring(startIndex, startIndex + 17);
        }
        return this.macAddress;
    }

    public void setNumVCPUs(final int numVCPUs) throws VMMException {
        try {
            this.domain.setVcpus(numVCPUs);
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#setSchedulingCap(int)
     */
    public void setSchedulingCap(final int schedulingCap) throws VMMException {
        SchedUintParameter cap = new SchedUintParameter(schedulingCap);
        cap.field = "cap";
        try {
            this.domain.setSchedulerParameters(new SchedParameter[] {cap});
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see
     * org.ow2.sirocco.vmm.api.VirtualMachineMXBean#setSchedulingWeight(int)
     */
    public void setSchedulingWeight(final int schedulingWeight) throws VMMException {
        SchedUintParameter cap = new SchedUintParameter(schedulingWeight);
        cap.field = "weight";
        try {
            this.domain.setSchedulerParameters(new SchedParameter[] {cap});
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see
     * org.ow2.sirocco.vmm.api.VirtualMachineMXBean#migrate(org.ow2.sirocco.
     * vmm.api.HostMXBean, boolean)
     */
    public void migrate(HostMXBean targetHost, final boolean live) throws IllegalOperationException, VMMException {
        this.setMigrating(true);
        try {
            String targetHostName = targetHost.getHostName();
            LibvirtVirtualMachine.logger.info("Attempting live migration of VM " + this.name + " to host " + targetHostName);

            boolean foundHostInServerPool = false;

            for (HostMXBean h : this.host.getServerPool().getManagedHosts()) {
                if (h.getHostName().equals(targetHost.getHostName())) {
                    foundHostInServerPool = true;
                    targetHost = h;
                    break;
                }
            }

            if (!foundHostInServerPool) {
                throw new IllegalOperationException("Source and target hosts belong to different server pools");
            }

            LibvirtHost targetLibvirtHost = (LibvirtHost) targetHost;

            LibvirtVirtualMachine.logger.info("VM " + this.name + " live migration to host " + targetHostName + "...");
            this.emitNotification(NotificationTypes.VM_MIGRATION_START, targetHostName, null);

            Domain newDomain = null;
            try {
                newDomain = this.domain.migrate(targetLibvirtHost.getLibvirtAPIConnection(), live ? 1 : 0, null, null, 0);
            } catch (LibvirtException ex) {
                LibvirtVirtualMachine.logger.error("Migration failure", ex);
                throw new VMMException("Migration failure: " + ex.getMessage());
            }

            try {
                String xmlDesc = this.domain.getXMLDesc(0);
                this.domain.undefine();
                newDomain.getConnect().domainDefineXML(xmlDesc);
            } catch (LibvirtException ex) {
                LibvirtVirtualMachine.logger.error(ex);
            }

            LibvirtVirtualMachine.logger.info("VM " + this.name + " live migration done");

            LibvirtHost oldHost = this.host;
            this.host.postMigrateVM(this, targetLibvirtHost);
            this.host = targetLibvirtHost;
            this.domain = newDomain;

            this.emitNotification(NotificationTypes.VM_MIGRATION, targetHostName, new String[] {
                oldHost.getObjectName().toString(), targetLibvirtHost.getObjectName().toString()});
        } finally {
            this.setMigrating(false);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#shutdown()
     */
    public void shutdown() throws VMMException {
        try {
            this.domain.shutdown();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#start()
     */
    public void start() throws VMMException {
        try {
            LibvirtVirtualMachine.logger.debug("START VM: uuid=" + this.domain.getUUIDString());
            this.domain.create();
            this.host.sync();
        } catch (LibvirtException ex) {
            // XXX
            LibvirtVirtualMachine.logger.error("Libvirt exception while starting VM " + this.name, ex);
            // throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#reboot()
     */
    public void reboot() throws VMMException {
        try {
            this.domain.reboot(0);
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#suspend()
     */
    public void suspend() throws VMMException {
        try {
            this.domain.save("/tmp/" + this.getNameLabel() + ".save");
            this.host.sync();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#resume()
     */
    public void resume() throws VMMException {
        try {
            this.domain.getConnect().restore("/tmp/" + this.getNameLabel() + ".save");
            this.host.sync();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public void pause() throws VMMException {
        try {
            this.domain.suspend();
            this.host.sync();
            // this.onVMStateChanged();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public void unpause() throws VMMException {
        try {
            this.domain.resume();
            this.host.sync();
            // this.onVMStateChanged();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    public void destroy() throws VMMException {
        this.destroy(true);
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.VirtualMachineMXBean#destroy()
     */
    public void destroy(final boolean deleteDisks) throws VMMException {
        super.removeFromResourcePartition();
        List<VirtualDisk> disks = this.getVirtualDisks();
        try {
            try {
                if (this.getState() == PowerState.RUNNING) {
                    this.domain.destroy();
                }
            } catch (Exception ex) {
            }
            if (deleteDisks) {
                for (VirtualDisk disk : this.getVirtualDisks()) {
                    try {
                        disk.getVolume().getStoragePool().destroyVolume(disk.getVolume().getKey());
                    } catch (VMMException ex) {
                        LibvirtVirtualMachine.logger.error("Failed to destroy volume " + disk.getVolume().getKey(), ex);
                    }
                }
            }
            this.domain.undefine();
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
        this.host.sync();
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] {new MBeanNotificationInfo(new String[] {NotificationTypes.VM_STATE_CHANGE,
            NotificationTypes.VM_MIGRATION_ABORT, NotificationTypes.VM_MIGRATION_START, NotificationTypes.VM_MIGRATION,
            NotificationTypes.LOG, NotificationTypes.ERROR}, Notification.class.getName(), "VM event")};
    }

    @Override
    public VirtualMachineImageMXBean makeTemplate(final String name, final String description,
        final Map<String, String> metadata) throws InsufficientResourcesException, IllegalOperationException,
        BadVMPowerStateException, VMMException {
        throw new UnsupportedOperationException();
    }

    @Override
    public synchronized List<VirtualDisk> getVirtualDisks() {
        if (this.disks == null) {
            this.disks = new ArrayList<VirtualDisk>();
            try {
                LibvirtDomainConfig domainConfig = LibvirtDomainConfig.loadDomainConfig(this.domain.getXMLDesc(0));
                List<Node> diskNodes = domainConfig.getDiskNodes("disk", "file");
                for (Node diskNode : diskNodes) {
                    String sourceFile = XMLUtils.getAttributeValue(diskNode, "source", "file");
                    String device = XMLUtils.getAttributeValue(diskNode, "target", "dev");
                    if (sourceFile != null && device != null) {
                        for (StoragePoolMXBean pool : this.host.serverPool.getStoragePools()) {
                            LibvirtStoragePool storagePool = (LibvirtStoragePool) pool;
                            if (sourceFile.contains(storagePool.getStoragePoolPath())) {
                                Volume vol = storagePool.getVolumeByKey(sourceFile);
                                if (vol != null) {
                                    VirtualDisk disk = new VirtualDisk();
                                    disk.setDevice(device);
                                    disk.setVolume(vol);
                                    this.disks.add(disk);
                                }
                            }
                        }

                    }
                }
            } catch (Exception ex) {
                LibvirtVirtualMachine.logger.error("Failed to get VM virtual disks", ex);
            }
        }
        return this.disks;
    }

    @Override
    public List<VirtualCdrom> getVirtualCdroms() {
        if (this.cdroms == null) {
            this.cdroms = new ArrayList<VirtualCdrom>();
            try {
                LibvirtDomainConfig domainConfig = LibvirtDomainConfig.loadDomainConfig(this.domain.getXMLDesc(0));
                List<Node> diskNodes = domainConfig.getDiskNodes("cdrom", "file");
                for (Node diskNode : diskNodes) {
                    String sourceFile = XMLUtils.getAttributeValue(diskNode, "source", "file");
                    String device = XMLUtils.getAttributeValue(diskNode, "target", "dev");
                    if (sourceFile != null) {
                        for (StoragePoolMXBean pool : this.host.serverPool.getStoragePools()) {
                            LibvirtStoragePool storagePool = (LibvirtStoragePool) pool;
                            if (sourceFile.contains(storagePool.getStoragePoolPath())) {
                                Volume vol = storagePool.getVolumeByKey(sourceFile);
                                if (vol != null) {
                                    VirtualCdrom cdrom = new VirtualCdrom();
                                    cdrom.setDevice(device);
                                    cdrom.setVolume(vol);
                                    this.cdroms.add(cdrom);
                                }
                            }
                        }
                    } else {
                        VirtualCdrom cdrom = new VirtualCdrom();
                        cdrom.setDevice(device);
                        cdrom.setVolume(null);
                        this.cdroms.add(cdrom);
                    }
                }
            } catch (Exception ex) {
                LibvirtVirtualMachine.logger.error("Failed to get VM virtual cdroms", ex);
            }
        }
        return this.cdroms;
    }

    static final int VIR_DOMAIN_DEVICE_MODIFY_LIVE = 1;

    static final int VIR_DOMAIN_DEVICE_MODIFY_CONFIG = 2;

    @Override
    public void attachVolume(final VirtualDiskSpec diskSpec) throws VMMException {
        Volume vol;
        String volName = "vol-" + this.name + "-" + System.nanoTime();
        if (diskSpec.getDiskOp() == DiskOperation.NEW) {
            vol = diskSpec.getStoragePool().createVolume(volName, diskSpec.getCapacityMB(), this.host.diskImageFormat);
        } else if (diskSpec.getDiskOp() == DiskOperation.CREATE_FROM) {
            vol = diskSpec.getStoragePool().createVolumeFrom(volName, diskSpec.getCapacityMB(), this.host.diskImageFormat,
                diskSpec.getVolume(), true);
        } else {
            vol = diskSpec.getVolume();
        }
        // find available device name
        String deviceNamePrefix = "hd";
        String deviceNameSuffix = "a";
        for (VirtualDisk disk : this.getVirtualDisks()) {
            deviceNamePrefix = disk.getDevice().substring(0, 2);
            String suffix = disk.getDevice().substring(2);
            if (suffix.compareTo(deviceNameSuffix) > 0) {
                deviceNameSuffix = suffix;
            }
        }
        deviceNameSuffix = deviceNameSuffix.replace(deviceNameSuffix.charAt(0), (char) (deviceNameSuffix.charAt(0) + 1));

        LibvirtDomainConfig config = null;

        try {
            config = LibvirtDomainConfig.newDomainConfig(this.host.serverPool.getHypervisor() + ".tmpl");
        } catch (Exception ex) {
            throw new VMMException("Internal error:" + ex.getMessage());
        }

        // String diskXml =
        // "<disk type='file' device='disk'><driver name='qemu' type='" +
        // diskSpec.getVolume().getFormat()
        // + "' cache='none'/> <source file='" + vol.getPath() +
        // "'/><target dev='" + deviceNamePrefix + deviceNameSuffix
        // + "' bus='virtio'/></disk>";

        String diskXml = config.getDiskDesc(vol.getPath(), deviceNamePrefix + deviceNameSuffix, diskSpec.getVolume()
            .getFormat());

        LibvirtVirtualMachine.logger.debug("Attaching volume " + diskXml);
        try {
            int flags = LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_LIVE;
            if (this.host.serverPool.getHypervisor().equals("xen")) {
                flags |= LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_CONFIG;
            }
            this.domain.attachDeviceFlags(diskXml, flags);
            VirtualDisk disk = new VirtualDisk();
            disk.setDevice(deviceNamePrefix + deviceNameSuffix);
            disk.setVolume(vol);
            this.disks.add(disk);
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        } catch (Throwable ex) {
            LibvirtVirtualMachine.logger.error("", ex);
        }
    }

    @Override
    public void detachVolume(final String key) throws VMMException {
        String deviceXml = null;
        try {
            LibvirtDomainConfig domainConfig = LibvirtDomainConfig.loadDomainConfig(this.domain.getXMLDesc(0));
            List<Node> diskNodes = domainConfig.getDiskNodes("disk", "file");
            for (Node diskNode : diskNodes) {
                if (key.equals(XMLUtils.getAttributeValue(diskNode, "source", "file"))) {
                    deviceXml = XMLUtils.xmlToString(diskNode);
                    break;
                }
            }
            if (deviceXml == null) {
                throw new VMMException("Cannot find volume " + key);
            }
            LibvirtVirtualMachine.logger.debug("Detaching volume " + deviceXml);
            int flags = LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_LIVE;
            if (this.host.serverPool.getHypervisor().equals("xen")) {
                flags |= LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_CONFIG;
            }
            this.domain.detachDeviceFlags(deviceXml, flags);
            for (VirtualDisk disk : this.getVirtualDisks()) {
                if (disk.getVolume().getKey().equals(key)) {
                    this.disks.remove(disk);
                    break;
                }
            }
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }
    }

    @Override
    public Volume makeImage(final String name, final StoragePoolMXBean storagePool) throws VMMException {
        if (this.getVirtualDisks().size() == 0) {
            throw new VMMException("VM " + name + " has no disk");
        }
        Volume parentVolume = this.getVirtualDisks().get(0).getVolume();
        return storagePool.createVolumeFrom(name + "-" + System.nanoTime(), parentVolume.getCapacityMB(),
            this.host.diskImageFormat, parentVolume, false);
    }

    @Override
    public void changeCdromMedia(final String device, final Volume iso) throws VMMException {
        String cdromXml = null;
        try {
            LibvirtDomainConfig domainConfig = LibvirtDomainConfig.loadDomainConfig(this.domain.getXMLDesc(0));
            List<Node> diskNodes = domainConfig.getDiskNodes("cdrom", "file");
            for (Node cdromNode : diskNodes) {
                if (device.equals(XMLUtils.getAttributeValue(cdromNode, "target", "dev"))) {
                    String path = iso == null ? "" : iso.getPath();
                    domainConfig.updateAttributeValue(cdromNode, "source", "file", path);
                    cdromXml = XMLUtils.xmlToString(cdromNode);
                    break;
                }
            }
        } catch (Exception ex) {
            LibvirtVirtualMachine.logger.error("while changing CD-ROM media", ex);
            throw new VMMException(ex.getMessage());
        }
        if (cdromXml == null) {
            throw new VMMException("Cannot find CD-ROM device" + device);
        }
        LibvirtVirtualMachine.logger.debug("Changing cdrom media:" + cdromXml);
        try {
            int flags = LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_LIVE;
            if (this.host.serverPool.getHypervisor().equals("xen")) {
                flags |= LibvirtVirtualMachine.VIR_DOMAIN_DEVICE_MODIFY_CONFIG;
            }
            this.domain.updateDeviceFlags(cdromXml, flags);
            for (VirtualCdrom cdrom : this.getVirtualCdroms()) {
                if (cdrom.getDevice().equals(device)) {
                    cdrom.setVolume(iso);
                    break;
                }
            }
        } catch (LibvirtException ex) {
            throw LibvirtVirtualMachine.translateLibvirtException(ex);
        }
    }

    @Override
    public List<VNIC> getVNICs() {
        // TODO Auto-generated method stub
        return null;
    }

    static VMMException translateLibvirtException(final LibvirtException ex) {
        // LibvirtVirtualMachine.logger.error("Libvirt exception:" +
        // ex.toString(), ex);
        return new VMMException(ex.toString());
    }
}
